import tensorflow as tf
import tensorflow.contrib as tf_contrib


class UtilityVariable:

    @staticmethod
    def initialize_weight(shape, l2_regular=('l2_regular', 0.0005), name=''):
        kernel = tf.truncated_normal(shape, stddev=0.1)
        if l2_regular[0] is not None:
            tf.add_to_collection(
                name=l2_regular[0],
                value=tf_contrib.layers.l2_regularizer(scale=l2_regular[1])(kernel)
            )
        return tf.Variable(kernel, name=name)

    @staticmethod
    def initialize_bias(shape, name=''):
        initial = tf.constant(0.1, shape=shape)
        return tf.Variable(initial, name=name)



import tensorflow as tf
import math


class UtilityLoss:
    @staticmethod
    def softmax_cross_entropy(batch_predict,
                              batch_truth,
                              batch_select,
                              keep_ratio,
                              name,
                              debug=False):
        """

        :param batch_predict:
        :param batch_truth: ground truth category. int32, int64 only
        :param batch_select: a list of selected categories.
        :param keep_ratio: keep top high 0.7 of cross entropy.
        :param name: a name for tf.summary
        :param debug: True: return loss,
                            valid_index,
                            raw_cross_entropy,
                            valid_cross_entropy,
                            top_cross_entropy;
                      False: return loss
        :return:
        """

        valid_condition = UtilityLoss._selected_index(
            batch_category=batch_truth,
            batch_select=batch_select
        )

        base_mask = tf.multiply(
            x=tf.ones_like(batch_truth),
            y=batch_select[0]
        )
        batch_truth = tf.cast(
            tf.where(
                condition=valid_condition,
                x=batch_truth,
                y=base_mask
            ),
            tf.int32
        )

        """
        cross entropy
        """
        raw_cross_entropy = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels=batch_truth,
            logits=batch_predict
        )

        zero_mask = tf.zeros_like(raw_cross_entropy)

        valid_cross_entropy = tf.where(
            condition=valid_condition,
            x=raw_cross_entropy,
            y=zero_mask
        )

        """
        keep ratio
        """
        valid_total = tf.reduce_sum(
            tf.cast(
                valid_condition,
                dtype=tf.float32)
        )

        keep_number = tf.cast(
            keep_ratio * valid_total,
            dtype=tf.int32
        )

        top_cross_entropy, _ = tf.nn.top_k(valid_cross_entropy, k=keep_number)

        """
        loss
        """
        loss = tf.reduce_mean(top_cross_entropy)

        tf.summary.scalar(name=name,
                          tensor=loss)

        if debug:
            return loss, valid_condition, raw_cross_entropy, valid_cross_entropy, top_cross_entropy
        else:
            return loss

    @staticmethod
    def l2(batch_predict,
           batch_truth,
           batch_category,
           batch_select,
           keep_ratio,
           name,
           debug=False):
        """

        :param batch_predict:
        :param batch_truth:
        :param batch_category: category label for each sample
        :param batch_select: a list of selected categories.
        :param keep_ratio: keep top high 0.7 of square error.
        :param name: a name for tf.summary
        :param debug: True: return loss,
                            valid_condition,
                            raw_square_error,
                            valid_square_error,
                            top_square_error;
                      False: return loss
        :return:
        """

        valid_condition = UtilityLoss._selected_index(
            batch_category=batch_category,
            batch_select=batch_select
        )

        """
        square error
        """
        raw_square_error = tf.reduce_sum(
            tf.square(batch_predict - batch_truth),
            axis=1
        )

        zero_mask = tf.zeros_like(raw_square_error)
        valid_square_error = tf.where(
            condition=valid_condition,
            x=raw_square_error,
            y=zero_mask
        )

        """
        keep ratio
        """
        valid_total = tf.reduce_sum(
            tf.cast(
                x=valid_condition,
                dtype=tf.float32)
        )
        keep_num = tf.cast(
            x=keep_ratio * valid_total,
            dtype=tf.int32
        )

        top_square_error, _ = tf.nn.top_k(valid_square_error, k=keep_num)

        """
        loss
        """
        loss = tf.reduce_mean(top_square_error)

        tf.summary.scalar(name=name,
                          tensor=loss)

        if debug:
            return loss, valid_condition, raw_square_error, valid_square_error, top_square_error
        else:
            return loss

    @staticmethod
    def arcface(x,
                weight,
                truth,
                margin,
                select,
                keep_ratio,
                name,
                s=64,
                debug=False):
        """

        :param x: [[x1],[x2],[x3]]
        :param weight: [[w11,w12,w13],[w21,w22,w23]]
        :param truth: [t1,t2,t3]
        :param margin:
        :param select:
        :param keep_ratio:
        :param name:
        :param s:
        :param debug:
        :return:
        """

        """
        valid index
        """
        category_number = len(select)
        selection_0 = select.pop(0)
        valid_index = tf.equal(truth, selection_0)

        for selection in select:
            valid_index = tf.logical_or(
                valid_index,
                tf.equal(truth, selection)
            )

        valid_total = tf.reduce_sum(
            tf.cast(
                valid_index,
                dtype=tf.float32)
        )

        """
        mask out no selected category
        """
        mask_b = tf.multiply(
            tf.ones_like(truth),
            selection_0
        )
        truth = tf.cast(
            tf.where(
                valid_index,
                truth,
                mask_b
            ),
            tf.int32
        )

        """
        add margin
        """
        cos_m = math.cos(margin)
        sin_m = math.sin(margin)

        x_norm = tf.nn.l2_normalize(x, 1)
        w_norm = tf.nn.l2_normalize(weight, 0)

        cos_theta = tf.matmul(x_norm, w_norm)
        sin_theta = tf.sqrt(
            1.0 - tf.square(cos_theta)
        )

        threshold = math.cos(math.pi - margin)
        cos_theta_m = tf.where(
            cos_theta >= threshold,
            cos_theta * cos_m - sin_theta * sin_m,
            cos_theta - margin * sin_m
        )

        truth_one_hot = tf.one_hot(indices=truth, depth=category_number)

        gap_cos_theta = tf.where(
            truth_one_hot == 1,
            cos_theta_m,
            cos_theta
        )

        gap_cos_theta *= s

        """
        cross entropy
        """
        raw_cross_entropy = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels=truth,
            logits=gap_cos_theta
        )

        zeros = tf.zeros_like(raw_cross_entropy)

        valid_cross_entropy = tf.where(
            valid_index,
            raw_cross_entropy,
            zeros
        )

        """
        keep ratio
        """
        keep_number = tf.cast(
            keep_ratio * valid_total,
            dtype=tf.int32
        )

        top_k_cross_entropy, _ = tf.nn.top_k(valid_cross_entropy, k=keep_number)

        """
        loss
        """
        loss = tf.reduce_mean(top_k_cross_entropy)

        tf.summary.scalar(name=name,
                          tensor=loss)

        if debug:
            return loss, valid_index, raw_cross_entropy, valid_cross_entropy, top_k_cross_entropy
        else:
            return loss

    @staticmethod
    def wing(batch_predict,
             batch_truth,
             batch_category,
             batch_select: list,
             w,
             e,
             keep_ratio,
             name):

        valid_condition = UtilityLoss._selected_index(
            batch_category=batch_category,
            batch_select=batch_select
        )

        """
        wing loss
        """
        abs_error = tf.abs(batch_truth - batch_predict)

        raw_loss = tf.reduce_sum(
            input_tensor=tf.where(
                condition=abs_error < w,
                x=w * tf.log(1 + abs_error / e),
                y=abs_error - (w - w * tf.log(1 + w / e))
            ),
            axis=1
        )

        zero_mask = tf.zeros_like(raw_loss)
        valid_loss = tf.where(
            condition=valid_condition,
            x=raw_loss,
            y=zero_mask
        )

        """
        get top k
        """
        valid_total = tf.reduce_sum(
            tf.cast(
                x=valid_condition,
                dtype=tf.float32)
        )
        keep_num = tf.cast(
            keep_ratio * valid_total,
            dtype=tf.int32
        )
        top_loss, _ = tf.nn.top_k(valid_loss, k=keep_num)

        """
        mean loss
        """
        loss = tf.reduce_mean(top_loss)

        tf.summary.scalar(name=name,
                          tensor=loss)

        return loss

    @staticmethod
    def _selected_index(batch_category, batch_select):

        _s = batch_select[0]

        valid_condition = tf.equal(
            x=batch_category,
            y=_s
        )

        for select in batch_select:
            valid_condition = tf.logical_or(
                x=valid_condition,
                y=tf.equal(
                    x=batch_category,
                    y=select
                )
            )

        return valid_condition


if __name__ == '__main__':
    from easydict import EasyDict

    predicted = EasyDict()
    predicted.category = tf.constant([[0.0, 1.0],
                                      [0.1, 0.9],
                                      [0.2, 0.8],
                                      [0.3, 0.7],
                                      [0.4, 0.6],
                                      [0.5, 0.5],
                                      [0.6, 0.4],
                                      [0.7, 0.3],
                                      [0.8, 0.2],
                                      [0.9, 0.1]])
    predicted.offset = tf.constant([[0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1]])

    ground_truth = EasyDict()
    ground_truth.category = tf.constant([1,
                                         -2,
                                         -2,
                                         1,
                                         -2,
                                         1,
                                         0,
                                         -2,
                                         0,
                                         1])
    ground_truth.offset = tf.constant([[0.0, 0.1, 0.1, 0.1],
                                       [0.0, 0.0, 0.1, 0.1],
                                       [0.0, 0.0, 0.0, 0.1],
                                       [0.0, 0.0, 0.0, 0.0],
                                       [0.3, 0.1, 0.1, 0.1],
                                       [0.3, 0.3, 0.1, 0.1],
                                       [0.3, 0.3, 0.3, 0.1],
                                       [0.3, 0.3, 0.3, 0.3],
                                       [0.1, 0.5, 0.1, 0.1],
                                       [0.1, 0.5, 0.1, 0.1]])

    cross_entropy_op = UtilityLoss.softmax_cross_entropy(batch_predict=predicted.category,
                                                         batch_truth=ground_truth.category,
                                                         batch_select=[0, 1],
                                                         keep_ratio=1,
                                                         name='category',
                                                         debug=False)

    offset_loss_op = UtilityLoss.l2(batch_predict=predicted.offset,
                                    batch_truth=ground_truth.offset,
                                    batch_category=ground_truth.category,
                                    batch_select=[2],
                                    keep_ratio=0.7,
                                    name='offset',
                                    debug=True)

    offset_wing_op = UtilityLoss.wing(batch_predict=predicted.offset,
                                      batch_truth=ground_truth.offset,
                                      batch_category=ground_truth.category,
                                      batch_select=[-2, 1],
                                      w=10,
                                      e=2,
                                      keep_ratio=0.7,
                                      name='offset')

    with tf.Session() as sess:
        cross_entropy = sess.run(cross_entropy_op)
        print(cross_entropy)

        offset_loss = sess.run(offset_loss_op)
        print(offset_loss)

        wing_loss = sess.run(offset_wing_op)
        print(wing_loss)



import tensorflow as tf


class UtilityActive:
    @staticmethod
    def prelu(features, name):
        alphas = tf.get_variable('%s/prelu_alpha' % name, features.get_shape()[-1],
                                 initializer=tf.constant_initializer(0.0),
                                 dtype=tf.float32)
        pos = tf.nn.relu(features)
        neg = alphas * (features - abs(features)) * 0.5

        return pos + neg



import time
import numpy
import cv2

import inspect

from easy_net_tf.utility.image import UtilityImage


class UtilityLog:
    ERROR = 'Error:   '
    LOG = 'Log:     '
    WARNING = 'Warning: '

    @staticmethod
    def text(content,
             name,
             tag,
             option):
        """

        :param content:
        :param name:
        :param tag: 'Error', 'Log', 'Warning'
        :param option: True: print content;
                    False: disabled;
                    string: save in it as file
        :return:
        """
        if option:
            content = '%s%s | [%s.%s] %s' % \
                      (tag,
                       time.strftime('%Y-%m-%d-%H:%M:%S'),
                       name,
                       inspect.stack()[1][3],
                       content)
            print(content)

            if type(option) is str and option is not '':
                with open(option, 'a') as file:
                    file.write(content + '\n')

        return

    @staticmethod
    def image(image,
              batch_rectangle=(
                      [
                          [1, 1, 1, 1],
                          [2, 2, 2, 2]
                      ],
                      (255, 0, 0)
              ),
              batch_point=(
                      [
                          [1, 1],
                          [2, 2]
                      ],
                      (0, 255, 0)
              ),
              text=(
                      'content',
                      (0, 0, 255)
              ),
              hint='window name',
              delay=0):

        if delay is None:
            return

        copy = numpy.copy(image)

        copy = UtilityImage.draw_rectangle(
            image=copy,
            batch_rectangle=batch_rectangle[0],
            color=batch_rectangle[1],
        )

        if batch_point is not None:
            copy = UtilityImage.draw_point(
                image=copy,
                batch_point=batch_point[0],
                color=batch_point[1]
            )

        copy = cv2.putText(
            img=copy,
            text=text[0],
            org=(0, round(image.shape[0] / 2)),
            fontFace=cv2.FONT_HERSHEY_SCRIPT_SIMPLEX,
            fontScale=3,
            color=text[1]
        )

        cv2.imshow(winname=hint, mat=copy)
        cv2.waitKey(delay)

        return



import numpy
import tensorflow as tf

from easy_net_tf.utility.variable import UtilityVariable


class MobileNet:
    VALID = 'VALID'
    SAME = 'SAME'

    def __init__(self,
                 batch_input,
                 dilate_number,
                 filter_size,
                 channels_out,
                 stride,
                 add_bias,
                 add_residual=False,
                 padding=SAME,
                 dilate_active_func=tf.nn.leaky_relu,
                 depth_active_func=tf.nn.leaky_relu,
                 l2_regular=('l2_regular', 0.0005),
                 name=''):
        """
        :param batch_input: a Tensor [batch, height, width, channel]
        :param dilate_number: if 'None', become MobileNet V1 without output activation
        :param filter_size:
        :param channels_out:
        :param stride:
        :param add_bias: True: add; False: not add
        :param add_residual: be sure input [height, width] = output [height, width]. stride=1 and padding='SAME'
        :param padding:
        :param dilate_active_func: active function for dilatation.
        :param depth_active_func: active function for depth-wise.
        :param name: layer name
        """

        """
        prepare
        """
        if batch_input is None:
            assert 'Error: [%s.%s] batch_image can not be ''None''.' % (MobileNet.__name__,
                                                                        MobileNet.__init__.__name__)
        else:
            _, _height_in, _width_in, _channels_in = batch_input.shape

            self._height_in = _height_in.value
            self._width_in = _width_in.value
            self._channels_in = _channels_in.value
            assert self._channels_in is not None, '[%s,%s] ' \
                                                  'the channels of input must be explicit, ' \
                                                  'otherwise filters can not be initialized.' % (MobileNet.__name__,
                                                                                                 MobileNet.__init__.__name__)

        self._channels_out = channels_out

        self._dilate_number = dilate_number

        self._filter_size = filter_size

        self._dilate_active_func = dilate_active_func
        self._dilate_active_func_name = 'None' \
            if self._dilate_active_func is None \
            else self._dilate_active_func.__name__

        self._depth_active_func = depth_active_func
        self._depth_active_func_name = 'None' \
            if self._depth_active_func is None \
            else self._depth_active_func.__name__

        self._l2_regular = l2_regular
        self._stride = stride

        if add_residual \
                and (stride != 1
                     or padding is not MobileNet.SAME):
            add_residual = False
            print('Error: '
                  '[%s.%s] '
                  'if residual added, '
                  'be sure stride = 1 and padding = %s.%s'
                  % (MobileNet.__name__,
                     MobileNet.__init__.__name__,
                     MobileNet.__name__,
                     MobileNet.SAME))
        self.add_residual = add_residual

        self.add_bias = add_bias
        self.padding = padding
        self.name = name

        """
        initialize filter
        """
        self.dilate_filter, \
        self.depth_filter, \
        self.compress_filter, \
        self.residual_filter, \
        self.bias = self._initialize_variable()

        """
        calculation
        """
        self.features = self._calculate(batch_input)

    def _initialize_variable(self):
        """
        initialize variable
        :return:
        """
        channel_multiplier = 1

        """
        dilatation_filter
        """
        channels_middle = self._channels_in \
            if self._dilate_number is None \
            else self._dilate_number

        dilatation_filter = None \
            if self._dilate_number is None \
            else UtilityVariable.initialize_weight(
            [1,
             1,
             self._channels_in,
             channels_middle],
            l2_regular=self._l2_regular,
            name='%s_mn/dilate' % self.name
        )

        """
        depth-wise_filter
        """
        depthwise_filter = UtilityVariable.initialize_weight(
            [self._filter_size,
             self._filter_size,
             channels_middle,
             channel_multiplier],
            l2_regular=self._l2_regular,
            name='%s_mn/depth' % self.name
        )

        """
        compress_filter
        """
        compress_filter = UtilityVariable.initialize_weight(
            [1,
             1,
             channels_middle * channel_multiplier,
             self._channels_out],
            l2_regular=self._l2_regular,
            name='%s_mn/compress' % self.name
        )

        """
        residual_filter
        """
        if self.add_residual and self._channels_in != self._channels_out:
            residual_filter = UtilityVariable.initialize_weight(
                [1,
                 1,
                 self._channels_in,
                 self._channels_out],
                l2_regular=self._l2_regular,
                name='%s_mn/residual' % self.name
            )
        else:
            residual_filter = None

        """
        bias
        """
        if self.add_bias:
            bias = UtilityVariable.initialize_bias(
                [self._channels_out],
                name='%s_mn/bias' % self.name
            )
        else:
            bias = None

        return dilatation_filter, \
               depthwise_filter, \
               compress_filter, \
               residual_filter, \
               bias

    def _calculate(self,
                   batch_input):
        """
        No activation function applied on output.
        :param batch_input:
        :return:
        """

        """
        dilate
        """
        if self.dilate_filter is not None:

            dilatation_o = tf.nn.conv2d(input=batch_input,
                                        filter=self.dilate_filter,
                                        strides=[1, 1, 1, 1],
                                        padding='SAME')

            if self._depth_active_func is not None:
                dilatation_o = self._dilate_active_func(
                    features=dilatation_o,
                    name='%s_mn/dilate_active' % self.name
                )
        else:
            dilatation_o = batch_input

        """
        depth-wise
        """
        convolution_o = tf.nn.depthwise_conv2d(input=dilatation_o,
                                               filter=self.depth_filter,
                                               strides=[1, self._stride, self._stride, 1],
                                               padding=self.padding)
        if self._depth_active_func is not None:
            convolution_o = self._depth_active_func(
                features=convolution_o,
                name='%s_mn/depth_active' % self.name
            )

        """
        compress
        """
        compress_o = tf.nn.conv2d(input=convolution_o,
                                  filter=self.compress_filter,
                                  strides=[1, 1, 1, 1],
                                  padding='SAME')

        """
        residual block
        """
        if self.add_residual is True:
            if self.residual_filter is not None:
                batch_input = tf.nn.conv2d(input=batch_input,
                                           filter=self.residual_filter,
                                           strides=[1, 1, 1, 1],
                                           padding='SAME')

            batch_output = tf.add(compress_o, batch_input)
        else:
            batch_output = compress_o

        """
        bias
        """
        if self.bias is not None:
            batch_output += self.bias

        return batch_output

    def get_features(self):
        """
        :return: image out
        """
        return self.features

    def get_variables(self, sess=None, save_dir=None):
        """
        :return: variable
        """
        if sess is None:
            return self.dilate_filter, \
                   self.depth_filter, \
                   self.compress_filter, \
                   self.residual_filter, \
                   self.bias
        else:
            dilate_filter = None if self.dilate_filter is None else sess.run(self.dilate_filter)
            depth_filter = None if self.depth_filter is None else sess.run(self.depth_filter)
            compress_filter = None if self.compress_filter is None else sess.run(self.compress_filter)
            residual_filter = None if self.residual_filter is None else sess.run(self.residual_filter)
            bias = None if self.bias is None else sess.run(self.bias)

            if save_dir is not None:
                if dilate_filter is not None:
                    shape = '[%d,%d,%d,%d]' % dilate_filter.shape
                    numpy.savetxt(fname='%s/mn-dilate-%s.txt' % (save_dir, shape),
                                  X=numpy.reshape(dilate_filter, [-1]),
                                  header='%s: dilatation filter' % MobileNet.__name__,
                                  footer='shape: %s' % shape)

                    numpy.save(file='%s/mn-dilate-%s.npy' % (save_dir, shape),
                               arr=dilate_filter)

                if depth_filter is not None:
                    shape = '[%d,%d,%d,%d]' % depth_filter.shape
                    numpy.savetxt(fname='%s/mn-depthwise-%s.txt' % (save_dir, shape),
                                  X=numpy.reshape(depth_filter, [-1]),
                                  header='%s: depthwise filter' % MobileNet.__name__,
                                  footer='shape: %s' % shape)

                    numpy.save(file='%s/mn-depthwise-%s.npy' % (save_dir, shape),
                               arr=depth_filter)

                if compress_filter is not None:
                    shape = '[%d,%d,%d,%d]' % compress_filter.shape
                    numpy.savetxt(fname='%s/mn-compress-%s.txt' % (save_dir, shape),
                                  X=numpy.reshape(compress_filter, [-1]),
                                  header='%s: compress filter' % MobileNet.__name__,
                                  footer='shape: %s' % shape)

                    numpy.save(file='%s/mn-compress-%s.npy' % (save_dir, shape),
                               arr=compress_filter)

                if residual_filter is not None:
                    shape = '[%d,%d,%d,%d]' % residual_filter.shape
                    numpy.savetxt(fname='%s/mn-residual-%s.txt' % (save_dir, shape),
                                  X=numpy.reshape(residual_filter, [-1]),
                                  header='%s: residual filter' % MobileNet.__name__,
                                  footer='shape: %s' % shape)

                    numpy.save(file='%s/mn-residual-%s.npy' % (save_dir, shape),
                               arr=residual_filter)

                if bias is not None:
                    shape = '[%d]' % bias.shape
                    numpy.savetxt(fname='%s/mn-bias-%s.txt' % (save_dir, shape),
                                  X=bias,
                                  header='%s: bias' % MobileNet.__name__,
                                  footer='shape: %s' % shape)

                    numpy.save(file='%s/mn-bias-%s.npy' % (save_dir, shape),
                               arr=bias)

            return dilate_filter, \
                   depth_filter, \
                   compress_filter, \
                   residual_filter, \
                   bias

    def get_config(self):
        """
        export config as a list
        :return:
        """

        """
        multiplication calculation
        """
        multiplication_dilate, \
        multiplication_depth, \
        multiplication_compress, \
        multiplication_residual = self.get_multiplication_amount(height_in=self._height_in,
                                                                 width_in=self._width_in,
                                                                 dilate_number=self._dilate_number,
                                                                 channels_in=self._channels_in,
                                                                 channels_out=self._channels_out,
                                                                 filter_size=self._filter_size,
                                                                 stride=self._stride,
                                                                 add_residual=self.add_residual,
                                                                 padding=self.padding)

        dilate_shape = None if self.dilate_filter is None else self.dilate_filter.shape
        depth_shape = None if self.depth_filter is None else self.depth_filter.shape
        compress_shape = None if self.compress_filter is None else self.compress_filter.shape
        residual_shape = None if self.residual_filter is None else self.residual_filter.shape
        bias_shape = None if self.bias is None else self.bias.shape

        config = [
            '\n### %s\n' % self.name,
            '> MobileNet\n',
            '- filter size: %d\n' % self._filter_size,
            '- channels in: %d\n' % self._channels_in,
            '- channels out: %d\n' % self._channels_out,
            '- dilatation: %s\n' % self._dilate_number,
            '- dilate active function: %s\n' % self._dilate_active_func_name,
            '- depth-wise active function: %s\n' % self._depth_active_func_name,
            '- stride: %d\n' % self._stride,
            '- add residual: %s\n' % self.add_residual,
            '- add bias: %s\n' % self.add_bias,
            '- padding: %s\n' % self.padding,
            '- variables:\n',
            '   - dilatation: %s\n' % dilate_shape,
            '   - depth-wise: %s\n' % depth_shape,
            '   - compress: %s\n' % compress_shape,
            '   - residual: %s\n' % residual_shape,
            '   - bias: %s\n' % bias_shape,
            '- multiplicative amount: %d\n' % (multiplication_dilate
                                               + multiplication_depth
                                               + multiplication_compress
                                               + multiplication_residual),
            '   - dilatation: %d\n' % multiplication_dilate,
            '   - depth-wise: %d\n' % multiplication_depth,
            '   - compress: %d\n' % multiplication_compress,
            '   - residual: %d\n' % multiplication_residual
        ]

        return config

    @staticmethod
    def get_multiplication_amount(height_in,
                                  width_in,
                                  dilate_number,
                                  channels_in,
                                  channels_out,
                                  filter_size,
                                  stride,
                                  add_residual,
                                  padding):

        if height_in is None or width_in is None:
            return -1, -1, -1, -1

        else:
            if padding == MobileNet.VALID:
                height_out = round((height_in - filter_size + 1) / stride)
                width_out = round((width_in - filter_size + 1) / stride)
            elif padding == MobileNet.SAME:
                height_out = round(height_in / stride)
                width_out = round(width_in / stride)
            else:
                height_out = 0
                width_out = 0
                print('Error: '
                      '[%s.%s] '
                      'wrong padding mode.' % (MobileNet.__name__,
                                               MobileNet.__init__.__name__))
            if dilate_number is None:
                dilate = 0
                depth = height_out * width_out * filter_size * filter_size * channels_in
                compress = height_out * width_out * channels_in * channels_out
            else:
                dilate = height_in * width_in * channels_in * dilate_number
                depth = height_out * width_out * filter_size * filter_size * dilate_number
                compress = height_out * width_out * dilate_number * channels_out

            if add_residual and channels_in != channels_out:
                residual = height_in * width_out * channels_in * channels_out
            else:
                residual = 0

            return dilate, depth, compress, residual


if __name__ == '__main__':
    from easy_net_tf.utility.file import UtilityFile
    from pathlib import Path

    image = numpy.array([[[[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]]]],
                        dtype=numpy.float32)

    image_ph = tf.placeholder(dtype=tf.float32, shape=[None, None, None, 3])

    mn = MobileNet(batch_input=image_ph,
                   dilate_number=1,
                   filter_size=3,
                   channels_out=15,
                   dilate_active_func=tf.nn.relu6,
                   depth_active_func=tf.nn.leaky_relu,
                   stride=1,
                   add_residual=True,
                   add_bias=False,
                   padding=MobileNet.SAME,
                   name='l1')

    UtilityFile.save_str_list('test-mn.md', mn.get_config())



import tkinter
import tkinter.filedialog
import tkinter.messagebox


class UtilityDialog:
    @staticmethod
    def file_dialog(func, hint='', initial_dir='', mode=None):
        r = tkinter.Tk()
        r.withdraw()

        if mode is None:
            result = func(title=hint,
                          initialdir=initial_dir)
        else:
            result = func(mode=mode,
                          title=hint,
                          initial_dir=initial_dir)

        return result

    @staticmethod
    def message(func, title, message):
        r = tkinter.Tk()
        r.withdraw()

        func(title=title,
             message=message)


if __name__ == '__main__':
    # path = UtilityDialog.file_dialog_tk_avoid(func=tkinter.filedialog.askdirectory,
    #                                           hint='hello')
    # print(path)

    UtilityDialog.message(
        func=tkinter.messagebox.showinfo,
        title='hello',
        message='message'
    )



import numpy
import cv2
import math

from easydict import EasyDict

from easy_net_tf.utility.file import UtilityFile
from easy_net_tf.utility.image import UtilityImage


class UtilityLandmark:

    @staticmethod
    def normalize_circle(landmark, batch_circle):
        """

        :param batch_circle: [batch, 3]
        :param landmark: an EasyDict object
                        landmark.left
                        landmark.center
                        landmark.right
        :return:
        """

        batch_normal_landmark = list()
        for circle in batch_circle:

            normal_landmark = EasyDict()
            normal_landmark.left = list()
            normal_landmark.center = list()
            normal_landmark.right = list()

            # left
            for x, y in numpy.reshape(
                    a=landmark.left,
                    newshape=[-1, 2]
            ):
                normal_landmark.left.append(
                    [(x - circle[0] + circle[2]) / (circle[2] * 2 + 1),
                     (y - circle[1] + circle[2]) / (circle[2] * 2 + 1)]
                )

            # center
            for x, y in numpy.reshape(
                    a=landmark.center,
                    newshape=[-1, 2]
            ):
                normal_landmark.center.append(
                    [(x - circle[0] + circle[2]) / (circle[2] * 2 + 1),
                     (y - circle[1] + circle[2]) / (circle[2] * 2 + 1)])

            # right
            for x, y in numpy.reshape(
                    a=landmark.right,
                    newshape=[-1, 2]
            ):
                normal_landmark.right.append(
                    [(x - circle[0] + circle[2]) / (circle[2] * 2 + 1),
                     (y - circle[1] + circle[2]) / (circle[2] * 2 + 1)])

            batch_normal_landmark.append(
                normal_landmark
            )

        return batch_normal_landmark

    @staticmethod
    def normalize_rectangle(landmark, batch_rectangle):
        """

        :param batch_rectangle: [batch, 4]
        :param landmark: an EasyDict object
                        landmark.left
                        landmark.center
                        landmark.right
        :return:
        """

        batch_normal_landmark = list()
        for rectangle in batch_rectangle:

            normal_landmark = EasyDict()
            normal_landmark.left = list()
            normal_landmark.center = list()
            normal_landmark.right = list()

            # left
            for x, y in numpy.reshape(
                    a=landmark.left,
                    newshape=[-1, 2]
            ):
                normal_landmark.left.append(
                    [(x - rectangle[0]) / (rectangle[2] - rectangle[0] + 1),
                     (y - rectangle[1]) / (rectangle[3] - rectangle[1] + 1)]
                )

            # center
            for x, y in numpy.reshape(
                    a=landmark.center,
                    newshape=[-1, 2]
            ):
                normal_landmark.center.append(
                    [(x - rectangle[0]) / (rectangle[2] - rectangle[0] + 1),
                     (y - rectangle[1]) / (rectangle[3] - rectangle[1] + 1)]
                )

            # right
            for x, y in numpy.reshape(
                    a=landmark.right,
                    newshape=[-1, 2]
            ):
                normal_landmark.right.append(
                    [(x - rectangle[0]) / (rectangle[2] - rectangle[0] + 1),
                     (y - rectangle[1]) / (rectangle[3] - rectangle[1] + 1)]
                )

            batch_normal_landmark.append(
                normal_landmark
            )

        return batch_normal_landmark

    @staticmethod
    def regress_circle(batch_circle,
                       batch_normal_landmark):
        """
        calibrate landmarks to real scale
        :param batch_circle: [batch, 3]
        :param batch_normal_landmark: [batch, n]
        :return:
        """

        copy_batch_landmark = numpy.copy(batch_normal_landmark)

        for landmark, circle in zip(copy_batch_landmark, batch_circle):

            size = circle[2] * 2

            for _, point in enumerate(numpy.reshape(
                    a=landmark,
                    newshape=[-1, 2]
            )):
                point[0] *= size
                point[0] += (circle[0] - circle[2])

                point[1] *= size
                point[1] += (circle[1] - circle[2])

        return copy_batch_landmark

    @staticmethod
    def regress_rectangle(batch_rectangle,
                          batch_normal_landmark):
        """
        calibrate landmarks to real scale
        :param batch_rectangle: [batch, 4]
        :param batch_normal_landmark: [batch, 10]
        :return:
        """

        copy_batch_landmark = numpy.copy(batch_normal_landmark)

        for landmark, rectangle in zip(copy_batch_landmark, batch_rectangle):

            width = rectangle[2] - rectangle[0] + 1
            height = rectangle[3] - rectangle[1] + 1

            for _, point in enumerate(numpy.reshape(
                    a=landmark,
                    newshape=[-1, 2]
            )):
                point[0] *= width
                point[0] += rectangle[0]

                point[1] *= height
                point[1] += rectangle[1]

        return copy_batch_landmark

    @staticmethod
    def align_face(image, landmarks, template):
        image_height, image_width, _ = image.shape

        landmarks = numpy.reshape(landmarks, [-1, 2])
        template = numpy.reshape(template, [-1, 2])

        # maybe it is unnecessary, wait for proving!!!
        for index, _ in enumerate(landmarks):
            landmarks[index][0] *= image_width
            landmarks[index][1] *= image_height

            template[index][0] *= image_width
            template[index][1] *= image_height

        mat = cv2.estimateRigidTransform(src=landmarks,
                                         dst=template,
                                         fullAffine=True)

        align_image = cv2.warpAffine(src=image,
                                     M=mat,
                                     dsize=(image_height, image_width))

        return align_image

    @staticmethod
    def horizontally_flip(batch_normal_landmark):

        batch_normal_landmark_flip = list()

        for normal_landmark in batch_normal_landmark:
            normal_landmark_flip = EasyDict()

            # left
            copy = numpy.copy(normal_landmark.left)
            copy = numpy.reshape(
                a=copy,
                newshape=[-1, 2]
            )
            copy[:, 0] = 1 - copy[:, 0]
            normal_landmark_flip.right = copy

            # center
            copy = numpy.copy(normal_landmark.center)
            copy = numpy.reshape(
                a=copy,
                newshape=[-1, 2]
            )
            copy[:, 0] = 1 - copy[:, 0]
            normal_landmark_flip.center = copy

            # right
            copy = numpy.copy(normal_landmark.right)
            copy = numpy.reshape(
                a=copy,
                newshape=[-1, 2]
            )
            copy[:, 0] = 1 - copy[:, 0]
            normal_landmark_flip.left = copy

            batch_normal_landmark_flip.append(
                normal_landmark_flip
            )

        return batch_normal_landmark_flip

    @staticmethod
    def _point_rotate(batch_point, center, angle):
        """

        :param batch_point: shape: [ 10 ]
        :param center:
        :param angle:
        :return:
        """

        copy = batch_point.copy()
        copy = numpy.reshape(
            numpy.array(copy,
                        dtype=numpy.float32),
            [-1, 2]
        )
        rotate_mat = cv2.getRotationMatrix2D(center, angle, 1)

        batch_point_out = list()
        for x, y in copy:
            batch_point_out.append(rotate_mat[0][0] * x + rotate_mat[0][1] * y + rotate_mat[0][2])
            batch_point_out.append(rotate_mat[1][0] * x + rotate_mat[1][1] * y + rotate_mat[1][2])

        return batch_point_out

    @staticmethod
    def rotate(landmark, center, angle):
        copy = EasyDict(landmark.copy())

        # left
        copy.left = UtilityLandmark._point_rotate(
            batch_point=copy.left,
            center=center,
            angle=angle
        )
        # center
        copy.center = UtilityLandmark._point_rotate(
            batch_point=copy.center,
            center=center,
            angle=angle
        )
        # right
        copy.right = UtilityLandmark._point_rotate(
            batch_point=copy.right,
            center=center,
            angle=angle
        )

        return copy

    @staticmethod
    def for_save(batch_landmark):
        batch_landmark_save = list()
        for landmark in batch_landmark:
            # left
            batch_landmark_save.extend(
                landmark.left
            )
            # center
            batch_landmark_save.extend(
                landmark.center
            )
            # right
            batch_landmark_save.extend(
                landmark.right
            )
        return batch_landmark_save

    @staticmethod
    def mmlab_2_circle(read_path,
                       write_path,
                       log=False):
        info_generator = UtilityFile.get_line_generator(
            read_path=read_path,
            shuffle=False
        )

        sample_index = 0
        with open(write_path, 'w') as file:
            while True:
                if log:
                    sample_index += 1
                    print('Log: '
                          '[%s.%s] '
                          'sample: %d' % (UtilityLandmark.__name__,
                                          UtilityLandmark.mmlab_2_circle.__name__,
                                          sample_index)
                          )
                try:
                    info = info_generator.__next__().strip().split(' ')
                    image_path = info.pop(0)
                    x_1 = int(info.pop(0))
                    x_2 = int(info.pop(0))
                    y_1 = int(info.pop(0))
                    y_2 = int(info.pop(0))

                    circle = [(x_1 + x_2) / 2,
                              (y_1 + y_2) / 2,
                              math.sqrt(
                                  math.pow(x_2 - x_1, 2) + math.pow(y_2 - y_1, 2)
                              ) / 2]

                    new_info = image_path + ';' \
                               + ' '.join(map(str, circle)) + ';' \
                               + ' '.join(map(str, info)) + '\n'

                    file.write(new_info)

                except StopIteration:
                    break

        return

    @staticmethod
    def verify_mmlab(read_path,
                     image_dir,
                     delay=500,
                     log=False):

        """
        verify and show original label
        :param read_path:
        :param image_dir:
        :param delay:
        :param log:
        :return:
        """
        info_generator = UtilityFile.get_line_generator(
            read_path=read_path,
            shuffle=False
        )

        sample_index = 0

        while True:
            if log:
                sample_index += 1
                print('Log: '
                      '[%s.%s] '
                      'sample: %d' % (UtilityLandmark.__name__,
                                      UtilityLandmark.mmlab_2_circle.__name__,
                                      sample_index)
                      )
            try:
                info = info_generator.__next__().strip().split(' ')
                image_path = info.pop(0)
                image = cv2.imread(image_dir + '/' + image_path)

                x_1 = int(info.pop(0))
                x_2 = int(info.pop(0))
                y_1 = int(info.pop(0))
                y_2 = int(info.pop(0))

                image = UtilityImage.draw_rectangle(image=image,
                                                    batch_rectangle=[[x_1, y_1, x_2, y_2]],
                                                    color=(0, 255, 0))

                landmark = numpy.reshape(
                    a=numpy.array(list(map(float, info)), dtype=numpy.int32),
                    newshape=[-1, 2]
                )
                image = UtilityImage.draw_point(image=image,
                                                batch_point=landmark,
                                                color=(0, 255, 0))

                cv2.imshow('Original Landmark verify',
                           image)
                cv2.waitKey(delay)
            except StopIteration:
                break

        return

    @staticmethod
    def verify_circle(read_path,
                      image_dir,
                      delay=500,
                      log=False):
        """
        verify circle format label
        :param read_path:
        :param image_dir:
        :param delay:
        :param log:
        :return:
        """
        info_generator = UtilityFile.get_line_generator(
            read_path=read_path,
            shuffle=False
        )

        sample_index = 0

        while True:
            if log:
                sample_index += 1
                print('Log: '
                      '[%s.%s] '
                      'sample: %d' % (UtilityLandmark.__name__,
                                      UtilityLandmark.mmlab_2_circle.__name__,
                                      sample_index)
                      )
            try:
                info = info_generator.__next__().strip().split(';')
                image_path = info.pop(0)
                image = cv2.imread(image_dir + '/' + image_path)

                circle = list(map(float, info.pop(0).split(' ')))

                image = UtilityImage.draw_circle(image=image,
                                                 batch_circle=[circle])

                landmark = numpy.reshape(
                    a=numpy.array(list(map(float, info.pop(0).split(' '))), dtype=numpy.int32),
                    newshape=[-1, 2]
                )
                image = UtilityImage.draw_point(image=image,
                                                batch_point=landmark,
                                                color=(0, 255, 0))

                cv2.imshow('Original Landmark verify',
                           image)
                cv2.waitKey(delay)
            except StopIteration:
                break

        return


if __name__ == '__main__':
    UtilityLandmark.mmlab_2_circle(
        read_path='/home/yehangyang/Documents/Gitlab/AI_database/mmlab/data_set/train/labels.txt',
        write_path='/home/yehangyang/Documents/Gitlab/AI_database/mmlab/data_set/train/labels-c.txt',
        log=True
    )

    # UtilityLandmark.verify_mmlab(
    #     read_path='/home/yehangyang/Documents/Gitlab/AI_database/mmlab/data_set/train/trainImageList.txt',
    #     image_dir='/home/yehangyang/Documents/Gitlab/AI_database/mmlab/data_set/image',
    #     log=True
    # )

    # UtilityLandmark.verify_circle(
    #     read_path='/home/yehangyang/Documents/Gitlab/AI_database/mmlab/data_set/train/labels-c.txt',
    #     image_dir='/home/yehangyang/Documents/Gitlab/AI_database/mmlab/data_set/image',
    #     log=True
    # )



import numpy
import tensorflow as tf
from easy_net_tf.utility.variable import UtilityVariable


class CNN:
    VALID = 'VALID'
    SAME = 'SAME'

    def __init__(self,
                 batch_input,
                 filter_size,
                 channels_out,
                 stride,
                 add_bias,
                 add_residual=False,
                 normalize=False,
                 padding=SAME,
                 l2_regular=('l2_regular', 0.0005),
                 name=''):
        """
        :param batch_input: a batch of feature map
        :param filter_size:
        :param channels_out:
        :param stride:
        :param add_bias: True: add; False: not add
        :param add_residual:
        :param normalize: weight normalize
        :param padding:
        :param name: layer name
        """

        """
        prepare
        """
        if batch_input is None:
            assert 'Error: [%s.%s] batch_image can not be ''None''.' % (CNN.__name__,
                                                                        CNN.__init__.__name__)
        else:
            _, _height_in, _width_in, _channels_in = batch_input.shape
            self._height_in = _height_in.value
            self._width_in = _width_in.value
            self._channels_in = _channels_in.value
            assert self._channels_in is not None, '[%s,%s] ' \
                                                  'the channels of input must be explicit, ' \
                                                  'otherwise filters can not be initialized.' % (CNN.__name__,
                                                                                                 CNN.__init__.__name__)

        self._filter_size = filter_size
        self._channels_out = channels_out
        self._stride = stride

        if add_residual \
                and (stride != 1
                     or padding is not CNN.SAME):
            add_residual = False
            print('Error: '
                  '[%s.%s] '
                  'if residual added, '
                  'be sure stride = 1 and padding = %s.%s'
                  % (CNN.__name__,
                     CNN.__init__.__name__,
                     CNN.__name__,
                     CNN.SAME))
        self._add_residual = add_residual
        self._add_bias = add_bias
        self._padding = padding
        self._l2_regular = l2_regular
        self._name = name

        """
        initialize variable
        """
        self._cnn_filter, \
        self._residual_filter, \
        self._cnn_bias = self._initialize_variable(name=name)

        """
        normalize
        """
        if normalize:
            self._cnn_filter = tf.nn.l2_normalize(self._cnn_filter,
                                                  2,
                                                  name='cnn_weight')

        """
        calculation
        """
        self._features = self._calculate(batch_input)

    def _initialize_variable(self, name):
        """
        initialize variable
        :return:
        """

        """
        cnn_filter
        """
        cnn_filter = UtilityVariable.initialize_weight(
            [self._filter_size,
             self._filter_size,
             self._channels_in,
             self._channels_out],
            l2_regular=self._l2_regular,
            name='%s_cnn/weight' % name
        )

        """
        cnn_bias
        """
        if self._add_bias:
            cnn_bias = UtilityVariable.initialize_bias(
                [self._channels_out],
                name='%s_cnn/bias' % name
            )
        else:
            cnn_bias = None

        """
        residual_filter
        """
        if self._add_residual and self._channels_in != self._channels_out:
            residual_filter = UtilityVariable.initialize_weight(
                [1,
                 1,
                 self._channels_in,
                 self._channels_out],
                l2_regular=self._l2_regular,
                name='%s_cnn/residual' % name
            )
        else:
            residual_filter = None

        return cnn_filter, residual_filter, cnn_bias

    def _calculate(self,
                   batch_input):
        """
        No activation function applied on output.
        :param batch_input:
        :return:
        """

        """
        convolution
        """
        convolution_o = tf.nn.conv2d(input=batch_input,
                                     filter=self._cnn_filter,
                                     strides=[1, self._stride, self._stride, 1],
                                     padding=self._padding)

        """
        residual block
        """
        if self._add_residual is True:
            if self._residual_filter is not None:
                batch_input = tf.nn.conv2d(input=batch_input,
                                           filter=self._residual_filter,
                                           strides=[1, 1, 1, 1],
                                           padding='SAME')

            batch_output = tf.add(convolution_o, batch_input)
        else:
            batch_output = convolution_o

        """
        bias
        """
        if self._cnn_bias is not None:
            batch_output += self._cnn_bias

        return batch_output

    def get_features(self):
        """
        :return:
        """
        return self._features

    def get_variables(self, sess=None, save_dir=None):
        """
        :return:
        """
        if sess is None:
            return self._cnn_filter, \
                   self._residual_filter, \
                   self._cnn_bias

        else:
            convolution_filter = None if self._cnn_filter is None else sess.run(self._cnn_filter)
            residual_filter = None if self._residual_filter is None else sess.run(self._residual_filter)
            bias = None if self._cnn_bias is None else sess.run(self._cnn_bias)

            if save_dir is not None:
                if convolution_filter is not None:
                    shape = '[%d,%d,%d,%d]' % convolution_filter.shape
                    numpy.savetxt(fname='%s/cnn-convolution-%s.txt' % (save_dir, shape),
                                  X=numpy.reshape(convolution_filter, [-1]),
                                  header='%s: convolution filter' % CNN.__name__,
                                  footer='shape: %s' % shape)

                    numpy.save(file='%s/cnn-convolution-%s.npy' % (save_dir, shape),
                               arr=convolution_filter)

                if residual_filter is not None:
                    shape = '[%d,%d,%d,%d]' % residual_filter.shape
                    numpy.savetxt(fname='%s/cnn-residual-%s.txt' % (save_dir, shape),
                                  X=numpy.reshape(residual_filter, [-1]),
                                  header='%s: residual filter' % CNN.__name__,
                                  footer='shape: %s' % shape)

                    numpy.save(file='%s/cnn-residual-%s.npy' % (save_dir, shape),
                               arr=residual_filter)

                if bias is not None:
                    shape = '[%d]' % bias.shape
                    numpy.savetxt(fname='%s/cnn-bias-%s.txt' % (save_dir, shape),
                                  X=bias,
                                  header='%s: bias' % CNN.__name__,
                                  footer='shape: %s' % shape)

                    numpy.save(file='%s/cnn-bias-%s.npy' % (save_dir, shape),
                               arr=bias)

            return convolution_filter, \
                   residual_filter, \
                   bias

    def get_config(self):

        multiplication_convolution, \
        multiplication_residual = self.get_multiplication_amount(height_in=self._height_in,
                                                                 width_in=self._width_in,
                                                                 filter_size=self._filter_size,
                                                                 stride=self._stride,
                                                                 channels_in=self._channels_in,
                                                                 channels_out=self._channels_out,
                                                                 padding=self._padding,
                                                                 add_residual=self._add_residual)

        filter_shape = None if self._cnn_filter is None else self._cnn_filter.shape
        residual_shape = None if self._residual_filter is None else self._residual_filter.shape
        bias_shape = None if self._cnn_bias is None else self._cnn_bias.shape

        config = [
            '\n### %s\n' % self._name,
            '- CNN\n',
            '- filter size: %d\n' % self._filter_size,
            '- channels in: %d\n' % self._channels_in,
            '- channels out: %d\n' % self._channels_out,
            '- stride: %d\n' % self._stride,
            '- add residual: %s\n' % self._add_residual,
            '- add bias: %s\n' % self._add_bias,
            '- padding: %s\n' % self._padding,
            '- variables:\n',
            '   - convolution: %s\n' % filter_shape,
            '   - bias: %s\n' % bias_shape,
            '   - residual: %s\n' % residual_shape,
            '- multiplicative amount: %d\n' % (multiplication_convolution + multiplication_residual),
            '   - convolution: %d\n' % multiplication_convolution,
            '   - residual: %d\n' % multiplication_residual
        ]
        return config

    @staticmethod
    def get_multiplication_amount(height_in,
                                  width_in,
                                  filter_size,
                                  stride,
                                  channels_in,
                                  channels_out,
                                  padding,
                                  add_residual):
        if height_in is None or width_in is None:
            return -1, -1

        else:
            if padding == CNN.SAME:
                height_out = round(height_in / stride)
                width_out = round(width_in / stride)
            elif padding == CNN.VALID:
                height_out = round((height_in - filter_size + 1) / stride)
                width_out = round((width_in - filter_size + 1) / stride)
            else:
                height_out = 0
                width_out = 0
                print('Error: '
                      '[%s.%s] '
                      'wrong padding mode.'
                      % (CNN.__name__,
                         CNN.__init__.__name__))

            convolution = height_out * width_out * filter_size * filter_size * channels_in * channels_out

            if add_residual and channels_in != channels_out:
                residual = height_in * width_out * channels_in * channels_out
            else:
                residual = 0

            return convolution, residual


if __name__ == '__main__':
    from easy_net_tf.utility.file import UtilityFile

    image = numpy.array([[[[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]]]],
                        dtype=numpy.float32)

    image_ph = tf.placeholder(dtype=tf.float32, shape=[None, None, None, 3])

    cnn = CNN(batch_input=image_ph,
              filter_size=2,
              channels_out=6,
              stride=1,
              add_bias=True,
              add_residual=True,
              normalize=True,
              padding=CNN.SAME,
              name='l1')

    cnn.get_features()

    UtilityFile.save_str_list('test-cnn.md', cnn.get_config())

    weight_op = tf.constant([[[[1, 1],
                               [2, 1],
                               [3, 1]],
                              [[3, 3],
                               [4, 2],
                               [3, 1]]
                              ],
                             [[[3, 1],
                               [2, 2],
                               [3, 1]],
                              [[3, 3],
                               [4, 2],
                               [3, 1]]
                              ]
                             ],
                            dtype=tf.float32)

    weight_op = tf.nn.l2_normalize(weight_op, 2)

    tem_op = tf.constant(
        [[1, 1, 1],
         [1, 2, 3],
         [2, 2, 2],
         [3, 3, 3]],
        dtype=tf.float32
    )
    tem_op = tf.nn.l2_normalize(tem_op, 1)

    with tf.Session() as sess:
        weight_final = sess.run(weight_op)
        tem_final = sess.run(tem_op)
        print(weight_final)



import os
import cv2
import numpy
import math
import zipfile

from pathlib import Path

from easy_net_tf.utility.bounding import UtilityBounding


class UtilityImage:
    NORMALIZE_DEFAULT = 'default'
    NORMALIZE_EACH_CHANNEL = 'normalize each channel to [low,high]'
    NORMALIZE_ALL_CHANNEL = 'normalize all channel to [low,high]'
    NORMALIZE_NONE = 'no normalization'
    LOW = 0
    HIGH = 1

    @staticmethod
    def normalize(image,
                  params=(LOW, HIGH, NORMALIZE_DEFAULT)):
        """

        :param image:
        :param params:
        :return:
        """

        low, high, mode = params
        copy_image = numpy.copy(image)

        if mode == UtilityImage.NORMALIZE_DEFAULT:
            copy_image = (copy_image - 127.5) / 128

        # normalize each channel separately to [-1, 1]
        elif mode == UtilityImage.NORMALIZE_EACH_CHANNEL:

            for channel in range(copy_image.shape[2]):
                maximum = numpy.max(copy_image[:, :, channel])
                minimum = numpy.min(copy_image[:, :, channel])
                copy_image[:, :, channel] = ((copy_image[:, :, channel] - minimum) /
                                             numpy.clip(a=maximum - minimum, a_min=1e-10, a_max=None)
                                             ) * (high - low) + low

        # normalize all channels together to [-1, 1]
        elif mode == UtilityImage.NORMALIZE_ALL_CHANNEL:
            maximum = numpy.max(copy_image)
            minimum = numpy.min(copy_image)
            copy_image = ((copy_image - minimum) / (numpy.clip(a=maximum - minimum, a_min=1e-10, a_max=None))
                          ) * (high - low) + low

        # no normalization
        elif mode == UtilityImage.NORMALIZE_NONE:
            pass

        return copy_image

    @staticmethod
    def resize(image, new_size):
        """

        :param image:
        :param new_size: (new_width, new_height)
        :return:
        """
        copy_image = numpy.copy(image)

        copy_image = cv2.resize(copy_image,
                                new_size,
                                interpolation=cv2.INTER_LINEAR)
        return copy_image

    @staticmethod
    def rotate(image, center, angle):
        """

        :param image:
        :param center: rotation center (x, y)
        :param angle:
        :return:
        """

        copy_image = numpy.copy(image)

        rotate_mat = cv2.getRotationMatrix2D(center, angle, 1)
        copy_image = cv2.warpAffine(copy_image, rotate_mat, (copy_image.shape[1], copy_image.shape[0]))

        return copy_image

    @staticmethod
    def horizontally_flip(image):
        """

        :param image:
        :return:
        """

        """
        image flip
        """
        if image is None:
            print('Error: '
                  '[UtilityImage.horizontally_flip] '
                  'image_in cant be None')
            return

        copy = numpy.copy(image)
        copy = cv2.flip(copy, 1)

        return copy

    @staticmethod
    def vertical_flip(image,
                      landmark_in):
        if image is None:
            return None, None
        copy = numpy.copy(image)
        copy = cv2.flip(copy, 0)

        for i in range(numpy.size(landmark_in)):
            if i % 2 == 1:
                # vertical flip
                landmark_in[i] = 1.0 - landmark_in[i]

        return copy, landmark_in

    @staticmethod
    def draw_rectangle(image,
                       batch_rectangle,
                       batch_tag=None,
                       color=(255, 0, 0),
                       font_style=(cv2.FONT_HERSHEY_SIMPLEX,
                                   1,
                                   1,
                                   True)):
        """
        draw bounding rectangles on image
        :param image:
        :param batch_rectangle: [x1, y1, x2, y2]
        :param batch_tag: a list of str
        :param color: (B, G, R)
        :param font_style: (fontFace, fontScale, thickness, if background)
        :return:
        """
        copy = numpy.copy(image)

        for index, rectangle in enumerate(batch_rectangle):
            x_1 = int(round(max(rectangle[0], 0)))
            y_1 = int(round(max(rectangle[1], 0)))
            x_2 = int(round(min(rectangle[2], copy.shape[1] - 1)))
            y_2 = int(round(min(rectangle[3], copy.shape[0] - 1)))

            cv2.rectangle(copy,
                          (x_1, y_1),
                          (x_2, y_2),
                          color,
                          2)

            if batch_tag is not None:
                copy = UtilityImage.draw_tag(
                    image=copy,
                    content=str(batch_tag[index]),
                    position=(x_2, y_2),
                    color=color,
                    style=font_style
                )

        return copy

    @staticmethod
    def draw_point(image,
                   batch_point,
                   color=(255, 0, 0)):
        """
        draw landmarks on image
        :param image:
        :param batch_point: shape:[None, 2]
        :param color:
        :return:
        """
        copy = numpy.copy(image)

        for x, y in batch_point:
            cv2.circle(img=copy,
                       center=(int(round(x)), int(round(y))),
                       radius=2,
                       color=color,
                       thickness=-1)

        return copy

    @staticmethod
    def draw_circle(image,
                    batch_circle,
                    batch_tag=None,
                    color=(255, 0, 0),
                    font_style=(cv2.FONT_HERSHEY_SIMPLEX,
                                1,
                                1,
                                True)):
        """

        :param image:
        :param batch_circle:
        :param batch_tag:
        :param color:
        :param font_style: (fontFace, fontScale, thickness, if background)
        :return:
        """
        copy = numpy.copy(image)

        for index, circle in enumerate(batch_circle):
            x, y, r = circle

            cv2.circle(
                img=copy,
                center=(int(round(x)),
                        int(round(y))),
                radius=int(round(r)),
                color=color
            )

            if batch_tag is not None:
                copy = UtilityImage.draw_tag(
                    image=copy,
                    content=str(batch_tag[index]),
                    position=(int(round(x + r / math.sqrt(2))),
                              int(round(y + r / math.sqrt(2)))),
                    color=color,
                    style=font_style
                )

        return copy

    @staticmethod
    def draw_tag(image,
                 content,
                 position,
                 color=(255, 0, 0),
                 style=(cv2.FONT_HERSHEY_SIMPLEX,
                        1,
                        1,
                        True)):
        """

        :param image:
        :param content:
        :param position:
        :param color:
        :param style: (fontFace, fontScale, thickness, if background)
        :return:
        """
        copy = numpy.copy(image)

        if style[3]:
            cv2.rectangle(
                img=copy,
                pt1=(position[0], position[1] - style[1] * 20),
                pt2=(position[0] + style[1] * 20 * len(content), position[1]),
                color=color,
                thickness=-1
            )

        cv2.putText(
            img=copy,
            text=str(content),
            org=position,
            fontFace=style[0],
            fontScale=style[1],
            thickness=style[2],
            color=(255, 255, 255) if style[3] else color
        )
        return copy

    @staticmethod
    def add_coordinate(batch_image,
                       add_r,
                       normalize,
                       debug=False):
        """

        :param batch_image: a batch of images
        :param add_r: r coordinate=sqrt((x-width/2)^2+(y-height/2)^2)
        :param normalize:
        :param debug:
        :return:
        """

        """
        prepare
        """
        copy_batch_image = numpy.copy(batch_image)

        """
        x coordinate
        """
        x_ones = numpy.ones(
            [copy_batch_image.shape[0], copy_batch_image.shape[1]],
            dtype=numpy.int32
        )

        x_ones = numpy.expand_dims(x_ones, -1)

        x_range = numpy.tile(
            numpy.expand_dims(
                numpy.arange(copy_batch_image.shape[2]),
                0
            ),
            [copy_batch_image.shape[0], 1]
        )

        x_range = numpy.expand_dims(x_range, 1)
        x_channel = numpy.matmul(x_ones, x_range)
        x_channel = numpy.expand_dims(x_channel, -1)

        """
        y coordinate
        """
        y_ones = numpy.ones(
            [copy_batch_image.shape[0], copy_batch_image.shape[2]],
            dtype=numpy.int32
        )
        y_ones = numpy.expand_dims(y_ones, 1)
        y_range = numpy.tile(
            numpy.expand_dims(
                numpy.arange(copy_batch_image.shape[1]),
                0
            ),
            [copy_batch_image.shape[0], 1]
        )
        y_range = numpy.expand_dims(y_range, -1)
        y_channel = numpy.matmul(y_range, y_ones)
        y_channel = numpy.expand_dims(y_channel, -1)

        """
        r coordinate
        """
        r_channel = numpy.sqrt(
            numpy.square(x_channel - (copy_batch_image.shape[2] - 1) / 2.0)
            + numpy.square(y_channel - (copy_batch_image.shape[1] - 1) / 2.0)
        )

        """
        normalize
        """
        if normalize:
            x_channel = x_channel / (copy_batch_image.shape[2] - 1.0) * 2.0 - 1.0
            y_channel = y_channel / (copy_batch_image.shape[1] - 1.0) * 2.0 - 1.0

            maximum = numpy.max(r_channel)
            minimum = numpy.min(r_channel)
            r_channel = ((r_channel - minimum) / (maximum - minimum)) * 2.0 - 1.0

        """
        concatenate
        """
        copy_batch_image = numpy.concatenate(
            (copy_batch_image, x_channel, y_channel),
            axis=-1
        )
        if add_r:
            if debug:
                print('Log: '
                      '[%s.%s] '
                      'add r coordinate.'
                      % (UtilityImage.__name__,
                         UtilityImage.add_coordinate.__name__))
            copy_batch_image = numpy.concatenate(
                (copy_batch_image, r_channel),
                axis=-1
            )

        return copy_batch_image

    VALID = 'VALID'
    SAME = 'SAME'

    @staticmethod
    def cut_out_circle(image,
                       batch_circle,
                       size_out):
        batch_square = UtilityBounding.circle_2_square(batch_circle)
        batch_image = UtilityImage.cut_out_rectangles(
            image=image,
            batch_rectangle=batch_square,
            size_out=size_out
        )

        return batch_image

    @staticmethod
    def cut_out_rectangle(image,
                          rectangle,
                          padding=VALID):
        """

        :param image:
        :param rectangle: [x1, y1, x2, y2]
        :param padding: 'VALID': output valid image area; 'SAME': padding '0' if rectangle exceeds image
        :return:
        """

        copy_image = numpy.copy(image)
        rectangle = numpy.round(rectangle)

        if padding == UtilityImage.VALID:
            image_out = copy_image[
                        int(rectangle[1]):int(rectangle[3] + 1),
                        int(rectangle[0]):int(rectangle[2] + 1),
                        :
                        ]

        elif padding == UtilityImage.SAME:

            x_1 = rectangle[0]
            y_1 = rectangle[1]
            x_2 = rectangle[2]
            y_2 = rectangle[3]

            m_height = y_2 - y_1 + 1
            m_width = x_2 - x_1 + 1

            if x_1 < 0:
                m_x_1 = -x_1
                x_1 = 0
            else:
                m_x_1 = 0

            if y_1 < 0:
                m_y_1 = -y_1
                y_1 = 0
            else:
                m_y_1 = 0

            if x_2 >= copy_image.shape[1]:
                m_x_2 = m_width - 1 - (x_2 - (copy_image.shape[1] - 1))
                x_2 = copy_image.shape[1] - 1
            else:
                m_x_2 = m_width - 1

            if y_2 >= copy_image.shape[0]:
                m_y_2 = m_height - 1 - (y_2 - (copy_image.shape[0] - 1))
                y_2 = copy_image.shape[0] - 1
            else:
                m_y_2 = m_height - 1

            # generate crop size mask
            image_out = numpy.zeros(
                shape=(int(m_height),
                       int(m_width),
                       int(copy_image.shape[2])),
                dtype=numpy.uint8
            )
            image_out[
            int(m_y_1):int(m_y_2) + 1,
            int(m_x_1):int(m_x_2) + 1,
            :
            ] = copy_image[
                int(y_1):int(y_2) + 1,
                int(x_1):int(x_2) + 1,
                :
                ]

        else:
            print('Error: '
                  '[%s.%s] '
                  'padding should be ''VALID'' or ''SAME'''
                  % (UtilityImage.__name__,
                     UtilityImage.cut_out_rectangle.__name__))
            image_out = None

        return image_out

    @staticmethod
    def cut_out_rectangles(image,
                           batch_rectangle,
                           size_out):
        """
        1.repair exceeding rectangles to valid rectangles, padding '0' in exceeding area
        2.resize to net input size;
        :param image:
        :param batch_rectangle:
        :param size_out: resize size
        :return: a batch of resize cropped image
        """

        copy_image = numpy.copy(image)

        batch_rectangle = numpy.around(batch_rectangle).astype(numpy.int32)
        rectangle_number = batch_rectangle.shape[0]

        batch_x_1 = batch_rectangle[:, 0]
        batch_y_1 = batch_rectangle[:, 1]
        batch_x_2 = batch_rectangle[:, 2]
        batch_y_2 = batch_rectangle[:, 3]

        batch_width = batch_x_2 - batch_x_1 + 1
        batch_height = batch_y_2 - batch_y_1 + 1

        c_x_1 = numpy.zeros(
            (batch_rectangle.shape[0],),
            numpy.int32
        )
        c_y_1 = numpy.zeros(
            (batch_rectangle.shape[0],),
            numpy.int32
        )
        c_x_2 = batch_width.copy() - 1
        c_y_2 = batch_height.copy() - 1

        index = numpy.where(numpy.greater(batch_x_2, copy_image.shape[1] - 1))[0]
        if numpy.size(index) > 0:
            c_x_2[index] = (batch_width[index] - 1) - (batch_x_2[index] - (copy_image.shape[1] - 1))
            batch_x_2[index] = copy_image.shape[1] - 1

        index = numpy.where(batch_y_2 > copy_image.shape[0] - 1)
        if numpy.size(index) > 0:
            c_y_2[index] = (batch_height[index] - 1) - (batch_y_2[index] - (copy_image.shape[0] - 1))
            batch_y_2[index] = copy_image.shape[0] - 1

        index = numpy.where(batch_x_1 < 0)
        if numpy.size(index) > 0:
            c_x_1[index] = 0 - batch_x_1[index]
            batch_x_1[index] = 0

        index = numpy.where(batch_y_1 < 0)
        if numpy.size(index) > 0:
            c_y_1[index] = 0 - batch_y_1[index]
            batch_y_1[index] = 0

        images_out = numpy.zeros((rectangle_number,
                                  size_out,
                                  size_out,
                                  copy_image.shape[2]),
                                 dtype=numpy.uint8)

        for i in range(rectangle_number):
            container = numpy.zeros((batch_height[i],
                                     batch_width[i],
                                     copy_image.shape[2]),
                                    dtype=numpy.uint8)
            container[
            c_y_1[i]:c_y_2[i] + 1,
            c_x_1[i]:c_x_2[i] + 1,
            :
            ] = copy_image[
                batch_y_1[i]:batch_y_2[i] + 1,
                batch_x_1[i]:batch_x_2[i] + 1,
                :
                ]

            images_out[i, :, :, :] = cv2.resize(
                container,
                (size_out, size_out)
            )

        return images_out

    @staticmethod
    def save_image(image,
                   category,
                   save_dir,
                   datum_from,
                   stride,
                   rotation=None,
                   landmarks=None,
                   offsets=None,
                   save_zip=True):
        """

        :param image:
        :param category: a list(negative:0, positive:1, partial:-1, landmark:-2).
        :param save_dir: directory of saving specific category
        :param datum_from: str
        :param stride: (folder save stride, label save stride)
        :param rotation:
        :param landmarks:
        :param offsets:
        :return:
        """

        """
        prepare
        """
        # check image
        if image is None:
            print('Error: [%s,%s] image can not be None'
                  % (UtilityImage.__name__,
                     UtilityImage.save_image.__name__))

        # get stride
        stride_folder, stride_label = stride

        # get image index
        save_dir.mkdir(parents=True, exist_ok=True)
        cache_path = save_dir / '.cache'

        if cache_path.exists():
            with open(cache_path.__str__(), 'r') as file:
                entity_index = int(file.readline().strip())
            with open(cache_path.__str__(), 'w') as file:
                file.write('%d\n' % (entity_index + 1))
        else:
            cache_path.parent.mkdir(parents=True, exist_ok=True)

            entity_index = 0
            with open(cache_path.__str__(), 'w') as file:
                file.write('%d\n' % (entity_index + 1))

        """
        save image
        """
        image_dir = save_dir / 'image' / datum_from
        image_name = '%d.jpg' % entity_index
        group = '%d' % int(entity_index / stride_folder)

        if save_zip:
            image_dir.mkdir(parents=True, exist_ok=True)
            zip_name = group + '.zip'

            cv2.imwrite(filename=(image_dir / 'buffer.jpg').__str__(),
                        img=image)

            with zipfile.ZipFile(
                    file=(image_dir / zip_name).__str__(),
                    mode='a'
            ) as image_zip:
                UtilityImage.save_zip_image(
                    zipfile_ob=image_zip,
                    path_for_save=(image_dir / 'buffer.jpg').__str__(),
                    path_in_zip=image_name,
                    delete=True
                )
        else:
            image_dir = image_dir / group
            image_dir.mkdir(parents=True, exist_ok=True)

            cv2.imwrite(filename=(image_dir / image_name).__str__(),
                        img=image)

        """
        save labels
        """
        label = ';' + ' '.join(map(str, category))

        if rotation is not None:
            label += ';' + ' '.join(map(str, rotation))

        if offsets is not None:
            label += ';' + ' '.join(map(str, offsets))

        if landmarks is not None:
            label += ';' + ' '.join(map(str, landmarks))

        content = datum_from + ' ' + group + ' ' + image_name + label + '\n'

        filename = Path('%d.txt' % int(entity_index / stride_label))
        label_path = save_dir / 'label' / filename
        label_path.parent.mkdir(parents=True, exist_ok=True)

        with label_path.open('a') as file:
            file.write(content)

        return entity_index

    @staticmethod
    def bytes2image(buffer):
        image = numpy.asarray(bytearray(buffer), dtype=numpy.uint8)
        image = cv2.imdecode(image, flags=cv2.IMREAD_COLOR)
        return image

    @staticmethod
    def read_zip_image(zipfile_ob, filename='image.jpg'):
        b = zipfile_ob.read(filename)
        image = UtilityImage.bytes2image(b)

        return image

    @staticmethod
    def save_zip_image(zipfile_ob,
                       path_for_save,
                       path_in_zip,
                       delete=False):
        zipfile_ob.write(
            filename=path_for_save,
            arcname=path_in_zip
        )

        if delete:
            os.remove(path_for_save)

        return


if __name__ == '__main__':
    img = numpy.array([1, 2, 3])

    for i in range(1000):
        UtilityImage.save_image(
            image=img,
            category=[0, 2],
            save_dir=Path('./delete'),
            datum_from='mmlab',
            stride=(5, 3),
            rotation=[19],
            landmarks=[0, 1, 4, 7.0, 8.9],
            offsets=[0.54, 0.89, 0.13],
            save_zip=True
        )
    # img = cv2.imread('/home/yehangyang/Desktop/crew/f_rgb_face_85100_34.jpg')
    #
    # img = UtilityImage.draw_rectangle(
    #     image=img,
    #     batch_rectangle=[
    #         [5, 5, 20, 20]
    #     ],
    #     batch_tag=[
    #         'hello'
    #     ],
    # )
    #
    # cv2.imshow('test', img)
    # cv2.waitKey(0)



import numpy
import tensorflow as tf
from easy_net_tf.utility.variable import UtilityVariable


class FCN:
    def __init__(self,
                 batch_input,
                 nodes_out,
                 add_bias=True,
                 normalize=False,
                 l2_regular=('l2_regular', 0.0005),
                 name=''):

        """

        :param batch_input: a Tensor [?, nodes_in]
        :param nodes_out: output nodes number
        :param add_bias:
        :param normalize:
        :param name: layer name
        """

        if batch_input is None:
            assert 'Error: [%s.%s] batch_image can not be ''None''.' % (FCN.__name__,
                                                                        FCN.__init__.__name__)
        else:
            _, _nodes_in = batch_input.shape
            self._nodes_in = _nodes_in.value
            assert self._nodes_in is not None, '[%s,%s] ' \
                                               'the dimension of input must be explicit, ' \
                                               'otherwise filters can not be initialized.' % (FCN.__name__,
                                                                                              FCN.__init__.__name__)

        self._nodes_out = nodes_out
        self._l2_regular = l2_regular
        self._name = name

        """
        initialize variable
        """
        self._weight, \
        self._bias = self._initialize_variable(
            add_bias=add_bias,
            name=name
        )

        """
        normalize variable
        """
        if normalize:
            self._weight = tf.nn.l2_normalize(self._weight,
                                              0,
                                              name='fcn_weight')

        """
        calculate
        """
        self._features = self._calculate(batch_input=batch_input)

    def _initialize_variable(self, add_bias, name=''):
        """

        :param name:
        :return:
        """
        weight = UtilityVariable.initialize_weight(
            shape=[self._nodes_in,
                   self._nodes_out],
            l2_regular=self._l2_regular,
            name='%s_fcn/weight' % name
        )

        bias = UtilityVariable.initialize_bias(
            [self._nodes_out],
            name='%s_fcn/bias' % name
        ) if add_bias else None

        return weight, bias

    def _calculate(self,
                   batch_input):
        """

        :param batch_input:
        :return: features
        """

        if self._bias is None:
            batch_output = tf.matmul(batch_input, self._weight)
        else:
            batch_output = tf.matmul(batch_input, self._weight) + self._bias

        return batch_output

    def get_features(self):
        """

        :return: features
        """
        return self._features

    def get_variables(self, sess=None, save_dir=None):
        """

        :return: weight, bias
        """

        if sess is None:
            return self._weight, self._bias
        else:
            weight = None if self._weight is None else sess.run(self._weight)
            bias = None if self._bias is None else sess.run(self._bias)

            if save_dir is not None:
                if weight is not None:
                    shape = '[%d,%d]' % weight.shape
                    numpy.savetxt(fname='%s/fcn-weight-%s.txt' % (save_dir, shape),
                                  X=weight,
                                  header='%s: weight' % FCN.__name__,
                                  footer='shape: %s' % shape)

                    numpy.save(file='%s/fcn-weight-%s.npy' % (save_dir, shape),
                               arr=weight)

                if bias is not None:
                    shape = '[%d]' % bias.shape
                    numpy.savetxt(fname='%s/fcn-bias-%s.txt' % (save_dir, shape),
                                  X=bias,
                                  header='%s: bias' % FCN.__name__,
                                  footer='shape: %s' % shape)

                    numpy.save(file='%s/fcn-bias-%s.npy' % (save_dir, shape),
                               arr=bias)

            return weight, bias

    def get_config(self):
        """
        export config as a list
        :return:
        """

        weight_shape = None if self._weight is None else self._weight.shape
        bias_shape = None if self._bias is None else self._bias.shape

        config = [
            '\n### %s\n' % self._name,
            '- Fully Connect Net\n',
            '- nodes in: %d\n' % self._nodes_in,
            '- nodes out: %d\n' % self._nodes_out,
            '- variables:\n',
            '   - weight: %s\n' % weight_shape,
            '   - bias: %s\n' % bias_shape,
            '- multiplicative amount: %d\n' % (self._nodes_in * self._nodes_out)
        ]
        return config


if __name__ == '__main__':
    from easy_net_tf.utility.file import UtilityFile
    from pathlib import Path

    image = numpy.array([[[[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                          [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]]]],
                        dtype=numpy.float32)

    image_ph = tf.placeholder(dtype=tf.float32, shape=[None, 60])

    fcn = FCN(batch_input=image_ph,
              nodes_out=30,
              add_bias=False,
              normalize=False,
              name='test')

    UtilityFile.save_str_list('test-fcn.md', fcn.get_config())



import tensorflow as tf


class UtilityAccuracy:

    @staticmethod
    def count(predict, truth):
        """
        binary classification only now
        :param predict:
        :param truth:
        :return:
        """

        predict = tf.cast(predict, dtype=tf.int32)

        truth = tf.cast(
            tf.reshape(truth, [-1]),
            dtype=tf.int32
        )

        mask_ones = tf.ones_like(truth, dtype=tf.int32)
        mask_zeros = tf.zeros_like(truth, dtype=tf.int32)

        tp = tf.reduce_sum(
            tf.cast(
                tf.logical_and(
                    tf.equal(truth, mask_ones),
                    tf.equal(predict, mask_ones)
                ),
                dtype=tf.float32))

        tn = tf.reduce_sum(
            tf.cast(
                tf.logical_and(
                    tf.equal(truth, mask_zeros),
                    tf.equal(predict, mask_zeros)
                ),
                dtype=tf.float32
            )
        )

        fp = tf.reduce_sum(
            tf.cast(
                tf.logical_and(
                    tf.equal(truth, mask_zeros),
                    tf.equal(predict, mask_ones)
                ),
                dtype=tf.float32
            )
        )

        fn = tf.reduce_sum(
            tf.cast(
                tf.logical_and(
                    tf.equal(truth, mask_ones),
                    tf.equal(predict, mask_zeros)
                ),
                dtype=tf.float32
            )
        )

        tf.summary.scalar(name='TP', tensor=tp)
        tf.summary.scalar(name='TN', tensor=tn)
        tf.summary.scalar(name='FP', tensor=fp)
        tf.summary.scalar(name='FN', tensor=fn)
        tf.summary.scalar(name='total', tensor=tp + tn + fp + fn)

        return tp, tn, fp, fn

    @staticmethod
    def recall(tp, fn):
        result = tp / (tp + fn)
        tf.summary.scalar(name='recall', tensor=result)
        return result

    @staticmethod
    def accuracy(tp, tn, fp, fn):
        result = (tp + tn) / (tp + tn + fp + fn)
        tf.summary.scalar(name='accuracy', tensor=result)
        return result

    @staticmethod
    def precision(tp, fp):
        result = tp / (tp + fp)
        tf.summary.scalar(name='precision', tensor=result)
        return result

    @staticmethod
    def f1_score(tp, fp, fn):
        result = 2 * tp / (2 * tp + fp + fn)
        tf.summary.scalar(name='f1_score', tensor=result)
        return result



import random
import os


class UtilityFile:
    @staticmethod
    def _line_generator(read_path):
        """
        get a line generator of a file
        :param read_path: string
        :return:
        """
        with open(read_path, 'r') as file:
            while True:
                line = file.readline()
                if line:
                    yield line
                else:
                    return

    @staticmethod
    def _line_shuffle(read_path):
        """
        first get all lines in file, then shuffle, finally save to file again.
        :param read_path: string
        :return:
        """
        # read
        with open(read_path, 'r') as file:
            info_list = file.readlines()

        # shuffle
        random.shuffle(info_list)

        # write
        with open(read_path, 'w') as file:
            for info in info_list:
                file.write(info)

    @staticmethod
    def get_file_list(directory, shuffle):
        """

        :param directory:
        :param shuffle:
        :return:
        """

        file_list = os.listdir(directory)

        if shuffle:
            random.shuffle(file_list)

        return file_list

    @staticmethod
    def get_line_generator(read_path,
                           shuffle):
        """

        :param read_path:
        :param shuffle: True: first shuffle and save, then return generator; False: return generator
        :return:
        """

        if shuffle:
            UtilityFile._line_shuffle(read_path)

        return UtilityFile._line_generator(read_path)

    @staticmethod
    def save_str_list(write_path,
                      info_list,
                      mode='w'):
        """
        save a list of info to a file
        :param write_path: string
        :param info_list: a list of info to save
        :param mode: file mode
        :return:
        """
        with open(write_path, mode) as file:
            for info in info_list:
                file.write(str(info))


if __name__ == '__main__':
    save_list = ['net',
                 0.88,
                 128,
                 'hello',
                 False,
                 True,
                 None]

    UtilityFile.save_str_list('test.txt', save_list)



import numpy
import cv2
import math

from easydict import EasyDict

from easy_net_tf.utility.file import UtilityFile


class UtilityBounding:
    CIRCLE = 'CIRCLE'
    RECTANGLE = 'RECTANGLE'

    @staticmethod
    def rectangle_2_square(batch_rectangle,
                           image_width,
                           image_height):
        """
        convert rectangles to square shape with avoiding exceeding original image by move bounding towards image
        :param batch_rectangle:
        :param image_width:
        :param image_height:
        :return:
        """

        if batch_rectangle is None:
            return None

        squares = numpy.copy(batch_rectangle)

        batch_height = batch_rectangle[:, 3] - batch_rectangle[:, 1] + 1
        batch_width = batch_rectangle[:, 2] - batch_rectangle[:, 0] + 1

        long_side = numpy.maximum(batch_height, batch_width)

        squares[:, 0] = batch_rectangle[:, 0] + 0.5 * batch_width - 0.5 * long_side
        squares[:, 1] = batch_rectangle[:, 1] + 0.5 * batch_height - 0.5 * long_side

        squares[:, 2] = squares[:, 0] + long_side - 1
        squares[:, 3] = squares[:, 1] + long_side - 1

        invalid_indexes = numpy.less(squares[:, 0], 0)
        squares[:, 2] = numpy.where(invalid_indexes,
                                    squares[:, 2] - squares[:, 0],
                                    squares[:, 2])
        squares[:, 0] = numpy.where(invalid_indexes,
                                    0,
                                    squares[:, 0])

        invalid_indexes = numpy.less(squares[:, 1], 0)
        squares[:, 3] = numpy.where(invalid_indexes,
                                    squares[:, 3] - squares[:, 1],
                                    squares[:, 3])
        squares[:, 1] = numpy.where(invalid_indexes,
                                    0,
                                    squares[:, 1])

        invalid_indexes = numpy.greater_equal(squares[:, 2], image_width)
        squares[:, 0] = numpy.where(invalid_indexes,
                                    squares[:, 0] - (squares[:, 2] - image_width + 1),
                                    squares[:, 0])
        squares[:, 2] = numpy.where(invalid_indexes,
                                    image_width - 1,
                                    squares[:, 2])

        invalid_indexes = numpy.greater_equal(squares[:, 3], image_height)
        squares[:, 1] = numpy.where(invalid_indexes,
                                    squares[:, 1] - (squares[:, 3] - image_height + 1),
                                    squares[:, 1])
        squares[:, 3] = numpy.where(invalid_indexes,
                                    image_height - 1,
                                    squares[:, 3])

        return squares

    @staticmethod
    def circle_2_square(batch_circle):
        """
        convert circle to square shape without avoiding exceeding original image by move bounding towards image
        :param batch_circle:
        :return:
        """

        batch_square = numpy.zeros(
            shape=(numpy.shape(batch_circle)[0], 4),
            dtype=numpy.int32
        )

        batch_square[:, 0] = batch_circle[:, 0] - batch_circle[:, 2]
        batch_square[:, 1] = batch_circle[:, 1] - batch_circle[:, 2]
        batch_square[:, 2] = batch_circle[:, 0] + batch_circle[:, 2]
        batch_square[:, 3] = batch_circle[:, 1] + batch_circle[:, 2]

        return batch_square

    @staticmethod
    def rectangle_offset_horizontally_flip(offset):
        """

        :param offset:[x1, y1, x2, y2]
        :return:
        """

        copy = offset.copy()
        copy[0], copy[2] = -copy[2], -copy[0]

        return copy

    @staticmethod
    def offset_horizontally_flip_circle(batch_offset):

        copy = numpy.copy(batch_offset)
        copy[:, 0] = -copy[:, 0]

        return copy

    @staticmethod
    def iou_circle(circle,
                   batch_circle):
        """
        calculate Intersection on Union of circle against to each of circles
        :param circle:
        :param batch_circle:
        :return: intersection on union
        """

        # prepare
        batch_circle = numpy.reshape(batch_circle, [-1, 3])
        pi = math.pi
        iou_list = list()

        # get distance between circles
        distances = numpy.sqrt(
            numpy.square(batch_circle[:, 0] - circle[0])
            + numpy.square(batch_circle[:, 1] - circle[1])
        )

        for index, circle_2 in enumerate(batch_circle):

            if circle_2[2] + circle[2] <= distances[index]:
                """
                separate
                """
                iou_list.append(0)

            elif numpy.abs(circle_2[2] - circle[2]) >= distances[index]:
                """
                include
                """
                if circle_2[2] > circle[2]:
                    g_radius = circle_2[2]
                    l_radius = circle[2]

                else:
                    g_radius = circle[2]
                    l_radius = circle_2[2]

                iou_list.append(
                    numpy.square(l_radius) / numpy.square(g_radius)
                )

            else:
                """
                intersect
                """
                # get rhombus_area
                perimeter = distances[index] + circle[2] + circle_2[2]

                rhombus_area = 2 * numpy.sqrt(
                    (perimeter / 2)
                    * (perimeter / 2 - circle_2[2])
                    * (perimeter / 2 - circle[2])
                    * (perimeter / 2 - distances[index])
                )

                # get sector area
                height = rhombus_area / distances[index]

                sector_area_1 = numpy.arcsin(height / circle_2[2]) * numpy.square(circle_2[2])
                sector_area_2 = numpy.arcsin(height / circle[2]) * numpy.square(circle[2])

                # get intersection
                inter_area = sector_area_1 + sector_area_2 - rhombus_area

                # get union
                union_area = pi * numpy.square(circle_2[2]) + pi * numpy.square(circle[2]) - inter_area

                # get iou
                iou = inter_area / union_area

                iou_list.append(iou)

        return iou_list

    @staticmethod
    def iou_circle2(circle,
                    batch_circle):
        """
        calculate as squares
        :param circle:
        :param batch_circle:
        :return:
        """

        batch_circle = numpy.reshape(batch_circle, [-1, 3])
        circle_area = (circle[2] * 2) * (circle[2] * 2)
        circles_area = (batch_circle[:, 2] * 2) * (batch_circle[:, 2] * 2)

        inter_x_1 = numpy.maximum(circle[0] - circle[2], batch_circle[:, 0] - batch_circle[:, 2])
        inter_y_1 = numpy.maximum(circle[1] - circle[2], batch_circle[:, 1] - batch_circle[:, 2])
        inter_x_2 = numpy.minimum(circle[0] + circle[2], batch_circle[:, 0] + batch_circle[:, 2])
        inter_y_2 = numpy.minimum(circle[1] + circle[2], batch_circle[:, 1] + batch_circle[:, 2])

        inter_w = numpy.maximum(0, inter_x_2 - inter_x_1 + 1)
        inter_h = numpy.maximum(0, inter_y_2 - inter_y_1 + 1)

        inter_area = inter_w * inter_h

        union = circle_area + circles_area - inter_area

        ious = inter_area / union

        return ious

    @staticmethod
    def iou_rectangle(rectangle,
                      batch_rectangle):
        """
        calculate Intersection on Union of rectangle against to each of rectangles
        :param rectangle:
        :param batch_rectangle:
        :return: ious, inter_area, rectangle_area, batch_rectangle_area
        """

        batch_rectangle = numpy.reshape(batch_rectangle, [-1, 4])

        rectangle_area = (rectangle[2] - rectangle[0] + 1) * (rectangle[3] - rectangle[1] + 1)
        batch_rectangle_area = (batch_rectangle[:, 2] - batch_rectangle[:, 0] + 1) * (
                batch_rectangle[:, 3] - batch_rectangle[:, 1] + 1)

        inter_x_1 = numpy.maximum(rectangle[0], batch_rectangle[:, 0])
        inter_y_1 = numpy.maximum(rectangle[1], batch_rectangle[:, 1])
        inter_x_2 = numpy.minimum(rectangle[2], batch_rectangle[:, 2])
        inter_y_2 = numpy.minimum(rectangle[3], batch_rectangle[:, 3])

        inter_w = numpy.maximum(0, inter_x_2 - inter_x_1 + 1)
        inter_h = numpy.maximum(0, inter_y_2 - inter_y_1 + 1)

        inter_area = inter_w * inter_h
        union = rectangle_area + batch_rectangle_area - inter_area  # + 10e-10

        ious = inter_area / union

        return ious, inter_area, rectangle_area, batch_rectangle_area

    UNION = 'Union'
    MIN = 'Minimum'

    @staticmethod
    def nms_rectangle(batch_rectangle,
                      batch_score,
                      threshold,
                      mode='Union'):
        """
        non maximum suppression
        :param batch_rectangle:
        :param batch_score:
        :param threshold:
        :param mode:
        :return: kept indexes
        """

        x_1 = batch_rectangle[:, 0]
        y_1 = batch_rectangle[:, 1]
        x_2 = batch_rectangle[:, 2]
        y_2 = batch_rectangle[:, 3]

        batch_rectangle_area = (x_2 - x_1 + 1) * (y_2 - y_1 + 1)
        increase_order_indexes = numpy.argsort(-batch_score, axis=0)

        kept_indexes = list()

        while numpy.size(increase_order_indexes) > 0:
            index = increase_order_indexes[0]
            other_indexes = increase_order_indexes[1:]

            kept_indexes.append(index)

            if numpy.size(other_indexes) <= 0:
                break

            # intersection over union
            inter_x_1 = numpy.maximum(x_1[index], x_1[other_indexes])
            inter_y_1 = numpy.maximum(y_1[index], y_1[other_indexes])
            inter_x_2 = numpy.minimum(x_2[index], x_2[other_indexes])
            inter_y_2 = numpy.minimum(y_2[index], y_2[other_indexes])

            inter_width = numpy.maximum(inter_x_2 - inter_x_1 + 1, 0.0)
            inter_height = numpy.maximum(inter_y_2 - inter_y_1 + 1, 0.0)

            inter_area = inter_width * inter_height

            if mode == UtilityBounding.UNION:
                ious = inter_area / \
                       (batch_rectangle_area[index]
                        + batch_rectangle_area[other_indexes]
                        - inter_area)
            elif mode == UtilityBounding.MIN:
                ious = inter_area / \
                       numpy.minimum(
                           batch_rectangle_area[index],
                           batch_rectangle_area[other_indexes]
                       )
            else:
                print('Error: '
                      '[%s.%s] '
                      'mns mode can only be ''%s'' or ''%s''.'
                      % (UtilityBounding.__name__,
                         UtilityBounding.nms_rectangle.__name__,
                         UtilityBounding.MIN,
                         UtilityBounding.UNION))
                return None

            mask_index = numpy.where(ious <= threshold)[0]
            if numpy.size(mask_index) <= 0:
                break
            increase_order_indexes = other_indexes[mask_index]

        return kept_indexes

    @staticmethod
    def nms_circle2(batch_circle,
                    batch_score,
                    threshold,
                    mode='Union'):
        """
        calculate iou in a square way
        non maximum suppression
        :param batch_circle:
        :param batch_score:
        :param threshold:
        :param mode:
        :return: kept indexes
        """

        x_1 = batch_circle[:, 0] - batch_circle[:, 2]
        y_1 = batch_circle[:, 1] - batch_circle[:, 2]
        x_2 = batch_circle[:, 0] + batch_circle[:, 2]
        y_2 = batch_circle[:, 1] + batch_circle[:, 2]

        batch_rectangle_area = (x_2 - x_1 + 1) * (y_2 - y_1 + 1)
        increase_order_indexes = numpy.argsort(-batch_score, axis=0)

        kept_indexes = list()

        while numpy.size(increase_order_indexes) > 0:
            index = increase_order_indexes[0]
            other_indexes = increase_order_indexes[1:]

            kept_indexes.append(index)

            if numpy.size(other_indexes) <= 0:
                break

            # intersection over union
            inter_x_1 = numpy.maximum(x_1[index], x_1[other_indexes])
            inter_y_1 = numpy.maximum(y_1[index], y_1[other_indexes])
            inter_x_2 = numpy.minimum(x_2[index], x_2[other_indexes])
            inter_y_2 = numpy.minimum(y_2[index], y_2[other_indexes])

            inter_width = numpy.maximum(inter_x_2 - inter_x_1 + 1, 0.0)
            inter_height = numpy.maximum(inter_y_2 - inter_y_1 + 1, 0.0)

            inter_area = inter_width * inter_height

            if mode == UtilityBounding.UNION:
                ious = inter_area / \
                       (batch_rectangle_area[index]
                        + batch_rectangle_area[other_indexes]
                        - inter_area)
            elif mode == UtilityBounding.MIN:
                ious = inter_area / \
                       numpy.minimum(
                           batch_rectangle_area[index],
                           batch_rectangle_area[other_indexes]
                       )
            else:
                print('Error: '
                      '[%s.%s] '
                      'mns mode can only be ''%s'' or ''%s''.'
                      % (UtilityBounding.__name__,
                         UtilityBounding.nms_rectangle.__name__,
                         UtilityBounding.MIN,
                         UtilityBounding.UNION))
                return None

            mask_index = numpy.where(ious <= threshold)[0]
            if numpy.size(mask_index) <= 0:
                break
            increase_order_indexes = other_indexes[mask_index]

        return kept_indexes

    @staticmethod
    def normalize_circle(circle, random):
        normal_offset = [
            (circle[0] - random[0]) / random[2],
            (circle[1] - random[1]) / random[2],
            (circle[2] - random[2]) / random[2]
        ]

        return normal_offset

    @staticmethod
    def normalize_rectangle(rectangle, random):

        if random[2] - random[0] == random[3] - random[1]:
            size = random[2] - random[0] + 1
        else:
            return False

        normal_offset = [
            (rectangle[0] - random[0]) / size,
            (rectangle[1] - random[1]) / size,
            (rectangle[2] - random[2]) / size,
            (rectangle[3] - random[3]) / size
        ]

        return normal_offset

    @staticmethod
    def regress_circle(batch_circle,
                       batch_normal_offset):
        """
        calibrate rectangles by adding offsets
        :param batch_circle:
        :param batch_normal_offset:
        :return:
        """

        regress_batch_circle = numpy.copy(batch_normal_offset)

        for circle, regress_circle in zip(batch_circle, regress_batch_circle):
            regress_circle *= circle[2]
            regress_circle += circle

        return regress_batch_circle

    @staticmethod
    def regress_rectangle(batch_rectangle,
                          batch_normal_offset):
        """
        calibrate rectangles by adding offsets
        :param batch_rectangle:
        :param batch_normal_offset:
        :return:
        """

        regress_batch_rectangle = numpy.copy(batch_normal_offset)

        for rectangle, regress_rectangle in zip(batch_rectangle, regress_batch_rectangle):
            width = rectangle[2] - rectangle[0] + 1
            height = rectangle[3] - rectangle[1] + 1

            regress_rectangle = numpy.reshape(
                a=regress_rectangle,
                newshape=[-1, 2]
            )

            regress_rectangle[:, 0] *= width
            regress_rectangle[:, 1] *= height

            regress_rectangle = numpy.reshape(
                a=regress_rectangle,
                newshape=[-1]
            )

            regress_rectangle += rectangle

        return regress_batch_rectangle

    @staticmethod
    def wilder_rectangle_2_circle(read_path,
                                  write_path,
                                  log=False):
        """

        :param read_path: str
        :param write_path: str
        :param log:
        :return:
        """
        info_generator = UtilityFile.get_line_generator(
            read_path=read_path,
            shuffle=False
        )

        sample_index = 0
        with open(write_path, 'w') as file:
            while True:
                if log:
                    sample_index += 1
                    print('Log: '
                          '[%s.%s] '
                          'sample: %d'
                          % (UtilityBounding.__name__,
                             UtilityBounding.wilder_rectangle_2_circle.__name__,
                             sample_index)
                          )
                try:
                    """
                    get info
                    """
                    info = info_generator.__next__().strip().split(' ')

                    image_path = info.pop(0)

                    batch_rectangle = numpy.reshape(
                        numpy.array(info, dtype=numpy.float32),
                        [-1, 4]
                    )

                    """
                    convert to circle
                    """
                    circle_list = list()
                    for rectangle in batch_rectangle:
                        center_x = (rectangle[0] + rectangle[2]) / 2
                        center_y = (rectangle[1] + rectangle[3]) / 2
                        radius = numpy.sqrt(
                            numpy.square(rectangle[0] - rectangle[2])
                            + numpy.square(rectangle[1] - rectangle[3])
                        ) / 2

                        circle = [center_x, center_y, radius]
                        circle_list.append(circle)

                    """
                    save to file
                    """
                    circles = numpy.reshape(
                        numpy.array(circle_list),
                        [-1]
                    )

                    new_info = image_path + ' ' + ' '.join(map(str, circles)) + '\n'
                    file.write(new_info)

                except StopIteration:
                    break
        return

    @staticmethod
    def verify_circle(read_path,
                      image_dir,
                      delay):
        """

        :param read_path: circle bounding label file path
        :param image_dir: image folder directory
        :param delay:
        :return:
        """
        info_generator = UtilityFile.get_line_generator(
            read_path=read_path,
            shuffle=False
        )

        while True:
            try:
                info = info_generator.__next__().strip().split(' ')
                image_name = '%s.jpg' % info.pop(0)
                image_path = image_dir + '/' + image_name

                image = cv2.imread(image_path)

                circle_list = numpy.reshape(
                    list(map(float, info)),
                    [-1, 3]
                )

                for circle in circle_list:
                    circle = numpy.round(circle)
                    x = int(circle[0])
                    y = int(circle[1])
                    radius = int(circle[2])

                    cv2.circle(img=image,
                               center=(x, y),
                               radius=radius,
                               color=(0, 255, 0),
                               thickness=2)

                cv2.imshow('circle bounding', image)
                cv2.waitKey(delay)

            except StopIteration:
                break


if __name__ == '__main__':
    UtilityBounding.wilder_rectangle_2_circle(
        read_path='/home/yehangyang/Documents/Gitlab/AI_database/wider_face/val/labels.txt',
        write_path='/home/yehangyang/Documents/Gitlab/AI_database/wider_face/val/labels-c.txt',
        log=True
    )



import os
from pathlib import Path
import tkinter
import tkinter.filedialog
import tkinter.dialog

import tensorflow as tf
from tensorflow.python import pywrap_tensorflow

from easy_net_tf.utility.dialog import UtilityDialog
from easy_net_tf.utility.file import UtilityFile
from easy_net_tf.utility.log import UtilityLog


class UtilityTf:
    @staticmethod
    def latest_checkpoint(model_dir=None,
                          initial_dir='',
                          hint=''):
        """

        :param model_dir:
        :param initial_dir:
        :param hint:
        :return:
        """
        if model_dir is None:

            model_dir = UtilityDialog.file_dialog(
                func=tkinter.filedialog.askdirectory,
                hint=hint,
                initial_dir=initial_dir
            )

            if len(model_dir) == 0:
                model_dir = ''

        return tf.train.latest_checkpoint(model_dir)

    @staticmethod
    def select_checkpoint(initial_dir='',
                          hint=''):
        path = UtilityDialog.file_dialog(
            tkinter.filedialog.askopenfilename,
            hint=hint,
            initial_dir=initial_dir
        )

        if len(path) == 0:
            return None
        else:
            path = Path(path)
            path = path.parent / path.stem

            return path.__str__()

    @staticmethod
    def get_variables(checkpoint_path):
        """

        :param checkpoint_path:
        :return:
        """
        reader = pywrap_tensorflow.NewCheckpointReader(checkpoint_path)
        var_to_shape_map = reader.get_variable_to_shape_map()

        var = dict()
        for key in var_to_shape_map:
            var['%s' % key] = reader.get_tensor(key)

        return var

    @staticmethod
    def save_checkpoint(sess,
                        new_merits,
                        save_dir,
                        net_name,
                        global_step,
                        number=5):
        """
        save a better checkpoint according to comparison of a list of merits
        :param sess:
        :param new_merits: a list of merits for comparing
        :param save_dir: Path format
        :param net_name:
        :param global_step:
        :param number: the number of model to be saved
        :return: True: a new model saved; False: new model is not saved
        """

        model_saver = tf.train.Saver()

        # .checkpoint not exist
        if (save_dir / '.checkpoint').exists() is False:

            # update .model_comparision file
            with (save_dir / '.checkpoint').open('w') as file:
                file.write('%s-%d;' % (net_name, global_step)
                           + ';'.join(map(str, new_merits))
                           + '\n')

            # save new model
            model_saver.save(
                sess=sess,
                save_path=(save_dir / net_name).__str__(),
                global_step=global_step
            )
            return True
        # .checkpoint exist
        else:
            with (save_dir / '.checkpoint').open('r') as file:
                caches = file.readlines()

            if len(caches) < number:

                # update .checkpoint file
                with (save_dir / '.checkpoint').open('a') as file:
                    file.write('%s-%d;' % (net_name, global_step)
                               + ';'.join(map(str, new_merits))
                               + '\n')

                # save new model
                model_saver.save(
                    sess=sess,
                    save_path=(save_dir / net_name).__str__(),
                    global_step=global_step
                )
                return True

            else:
                for cache_index, cache in enumerate(caches):
                    records = cache.strip().split(';')
                    old_name = records.pop(0)
                    old_merits = list(map(float, records))

                    """
                    if all new merits greater than old merits
                    """
                    replace = True
                    for i, old_merit in enumerate(old_merits):
                        if new_merits[i] < old_merit:
                            replace = False
                            break

                    if replace:
                        """
                        update .checkpoint file
                        """
                        caches[cache_index] = '%s-%d;' % (net_name, global_step) \
                                              + ';'.join(map(str, new_merits)) \
                                              + '\n'
                        UtilityFile.save_str_list(
                            write_path=(save_dir / '.checkpoint').__str__(),
                            info_list=caches,
                            mode='w'
                        )

                        """
                        delete records file
                        """
                        file_list = UtilityFile.get_file_list(
                            directory=save_dir.__str__(),
                            shuffle=False
                        )
                        for filename in file_list:
                            if filename.find(old_name) >= 0:
                                os.remove((save_dir / filename).__str__())

                        """
                        save new model
                        """
                        model_saver.save(
                            sess=sess,
                            save_path=(save_dir / net_name).__str__(),
                            global_step=global_step
                        )

                        return True

                return False

    @staticmethod
    def restore_model(sess,
                      checkpoint_path,
                      name,
                      log=False):
        """

        :param sess:
        :param checkpoint_path:
        :param name: net name
        :param log: True: print restored model path; False: not
        :return: global_step
        """
        model_saver = tf.train.Saver()
        model_saver.restore(sess=sess,
                            save_path=checkpoint_path)
        global_step = int(checkpoint_path.split('/%s-' % name).pop())

        UtilityLog.text(
            content='restore model:\n'
                    '    %s' % checkpoint_path,
            name=UtilityTf.__name__,
            tag=UtilityLog.LOG,
            option=log
        )

        return global_step


if __name__ == '__main__':
    # path = UtilityTf.latest_checkpoint(hint='hello world')

    cp_path = UtilityTf.select_checkpoint(hint='Select a checkpoint file')
    variables = UtilityTf.get_variables(checkpoint_path=cp_path)

    print(variables)
